@@ -17,14 +17,11 @@ pub enum ValidationResult {
 
 impl ValidationResult {
     pub(crate) fn is_valid(&self) -> bool {
-        matches!(self, ValidationResult::Valid(_))
+        matches!(self, Self::Valid(_))
     }
 
     pub(crate) fn has_message(&self) -> bool {
-        matches!(
-            self,
-            ValidationResult::Valid(Some(_)) | ValidationResult::Invalid(Some(_))
-        )
+        matches!(self, Self::Valid(Some(_)) | Self::Invalid(Some(_)))
     }
 }
 
@@ -51,10 +48,11 @@ impl<'i> ValidationContext<'i> {
 }
 
 /// This trait provides an extension interface for determining whether
-/// the current input buffer is valid. Rustyline uses the method
-/// provided by this trait to decide whether hitting the enter key
-/// will end the current editing session and return the current line
-/// buffer to the caller of `Editor::readline` or variants.
+/// the current input buffer is valid.
+///
+/// Rustyline uses the method provided by this trait to decide whether hitting
+/// the enter key will end the current editing session and return the current
+/// line buffer to the caller of [`crate::Editor::readline`] or variants.
 pub trait Validator {
     /// Takes the currently edited `input` and returns a
     /// `ValidationResult` indicating whether it is valid or not along
@@ -64,8 +62,8 @@ pub trait Validator {
     /// delimiters are fully balanced.
     ///
     /// If you implement more complex validation checks it's probably
-    /// a good idea to also implement a `Hinter` to provide feedback
-    /// about what is invalid.
+    /// a good idea to also implement a [`crate::hint::Hinter`] to provide
+    /// feedback about what is invalid.
     ///
     /// For auto-correction like a missing closing quote or to reject invalid
     /// char while typing, the input will be mutable (TODO).
@@ -88,16 +86,6 @@ pub trait Validator {
 
 impl Validator for () {}
 
-impl<'v, V: ?Sized + Validator> Validator for &'v V {
-    fn validate(&self, ctx: &mut ValidationContext) -> Result<ValidationResult> {
-        (**self).validate(ctx)
-    }
-
-    fn validate_while_typing(&self) -> bool {
-        (**self).validate_while_typing()
-    }
-}
-
 /// Simple matching bracket validator.
 #[derive(Default)]
 pub struct MatchingBracketValidator {
